#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <fcntl.h>
#include "hdfi.h"
#include "cris.h"
#include "sis.h"

#include "porting_defines.h"
#include "crisSisGetUtilities.h"
#include "utilities.h"

    /* procedure to use in case of error (defined below) */
static void localTerminationProcedure(char *errorMessage);

	/* local helper routines */
static int zeroLastStructIfValidAtStopTime(char *pStructs, char *structName,
				uint32 stopTime, int nonzeroStructCount);
static int countNonzeroStructsInArray(char *pStructs, char *structName);
static uint32 getWorkingStartTime(uint32 secondsSince1996, char *structName);
static void *getSubsetOfDaysStructs(int day1996, char *path, char *structName,
				uint32 getStampsAfter, uint32 getStampsUpTo);
static char *integrateOverlapStructs(char *structName, char *pStructs,
			int nonOverlapStructCount, char *pOverlapStructs,
			int overlapStructCount);


void *getStructs(uint32 secondsSince1996Start, uint32 secondsSince1996Stop,
							char *structName) {
    char path[256];
    struct tm *pTimeStruct;
    char instrumentID, zero = 0;
    char dataFileSuffix[10];
    int day1996Start, day1996Stop, day1996;
    int maxStructCount, structSize, hasHeader, structCount, overlapStructCount;
    int nonzeroStructCount;
    char *pTemp_0, *pTemp, *pStructBuffer_0, *pStructBuffer;
    uint32 getStampsAfter, prelimGetStampsUpTo, timeStamp;

    void *pOverlapStructs_0, *pOverlapStructs;

    structSize = getStructSizeFromName(structName);

	/* first get overlap structs */
    pOverlapStructs_0 = getOverlapStructs(secondsSince1996Start,
					secondsSince1996Stop, structName);
    pOverlapStructs = pOverlapStructs_0;
    timeStamp = getTimeStamp(pOverlapStructs, structName);
    overlapStructCount = 0;
    while (timeStamp > 0) {
	overlapStructCount++;
	pOverlapStructs += structSize;
	timeStamp = getTimeStamp(pOverlapStructs, structName);
    }

	/* get standard archive file's structs */
    getStampsAfter = getWorkingStartTime(secondsSince1996Start, structName) - 1;
    prelimGetStampsUpTo = secondsSince1996Stop;	/* last struct chopped later? */

    instrumentID = getInstrumentIDFromStructName(structName);
    getDataFileSuffixFromStructName(dataFileSuffix, structName);

    day1996Start = getDay1996FromSecondsSince1996(getStampsAfter + 1);
    day1996Stop = getDay1996FromSecondsSince1996(prelimGetStampsUpTo);

	/* easy case.  start and stop in same day. */
    if (day1996Start == day1996Stop) {
	char *pStructs;

	pTimeStruct = getDateFromDay1996(day1996Start);
	generateDataFilePath(path, pTimeStruct, instrumentID, dataFileSuffix);

	pStructs = getSubsetOfDaysStructs(day1996Start, path, structName,
					getStampsAfter, prelimGetStampsUpTo);

	    /* if array holds more than 1 non-zero struct, chop off the */
	    /* last non-zero struct if it is valid at secondsSince1996Stop. */
	    /* this is to prevent the return of overlapping data on a */
	    /* series of calls to getStructs. */
	nonzeroStructCount = countNonzeroStructsInArray(pStructs, structName);
	if (nonzeroStructCount > 1) {
	    int zeroedStructCount;

	    zeroedStructCount = zeroLastStructIfValidAtStopTime(pStructs,
			structName, secondsSince1996Stop, nonzeroStructCount);
	    if (zeroedStructCount > 0) {
		nonzeroStructCount -= zeroedStructCount;
	    }
	}

	if (overlapStructCount > 0) {
	    pOverlapStructs = pOverlapStructs_0;
	    pStructs = integrateOverlapStructs(structName, pStructs,
		nonzeroStructCount, pOverlapStructs, overlapStructCount);
	}
	free(pOverlapStructs_0);

	return pStructs;
    }

	/* more general case, get structs across multiple days... */

	/* allocate enough memory to put all incoming structs */
    hasHeader = 1;
    maxStructCount = 0;
    for (day1996 = day1996Start; day1996 <= day1996Stop; day1996++) {
	pTimeStruct = getDateFromDay1996(day1996);
	generateDataFilePath(path, pTimeStruct, instrumentID, dataFileSuffix);
	maxStructCount += getFilesStructCount(path, structSize, hasHeader);
    }
    maxStructCount += 1;	/* terminal struct */
    pStructBuffer_0 = (char *) calloc(maxStructCount, structSize);
    if (pStructBuffer_0 == NULL) {
	localTerminationProcedure(
	    "ERROR: calloc failed.  probably not enough memory for allocation."
	);
    }
    pStructBuffer = pStructBuffer_0;

    structCount = 0;

	/* deal with partial start day */
    pTimeStruct = getDateFromDay1996(day1996Start);
    generateDataFilePath(path, pTimeStruct, instrumentID, dataFileSuffix);
    pTemp_0 = (char *) getSubsetOfDaysStructs(day1996Start, path, structName,
					getStampsAfter, prelimGetStampsUpTo);
    pTemp = pTemp_0;
    while (getTimeStamp(pTemp, structName)) {
	memcpy(pStructBuffer, pTemp, structSize);
	pTemp += structSize;
	pStructBuffer += structSize;
	structCount++;
    }
    free(pTemp_0);

	/* deal with any full days between day1996Start and day1996Stop */

    for (day1996 = day1996Start + 1; day1996 < day1996Stop; day1996++) {
	pTimeStruct = getDateFromDay1996(day1996);
	generateDataFilePath(path, pTimeStruct, instrumentID, dataFileSuffix);
	pTemp_0 = (char *) getSubsetOfDaysStructs(day1996, path, structName,
					getStampsAfter, prelimGetStampsUpTo);
	pTemp = pTemp_0;
	while (getTimeStamp(pTemp, structName)) {
	    memcpy(pStructBuffer, pTemp, structSize);
	    pTemp += structSize;
	    pStructBuffer += structSize;
	    structCount++;
	}
	free(pTemp_0);
    }

	/* deal with partial stop day */
    pTimeStruct = getDateFromDay1996(day1996Stop);
    generateDataFilePath(path, pTimeStruct, instrumentID, dataFileSuffix);
    pTemp_0 = (char *) getSubsetOfDaysStructs(day1996Stop, path, structName,
					getStampsAfter, prelimGetStampsUpTo);
    pTemp = pTemp_0;
    while (getTimeStamp(pTemp, structName)) {
	memcpy(pStructBuffer, pTemp, structSize);
	pTemp += structSize;
	pStructBuffer += structSize;
	structCount++;
    }
    free(pTemp_0);

	/* terminal zeroed struct */
    memset(pStructBuffer, zero, structSize);
    pStructBuffer += structSize;	/* reminder only */
    structCount++;

	/* resize allocated memory down to fit */
    pStructBuffer_0 =
	(char *) realloc(pStructBuffer_0, (structSize * structCount));
    if (pStructBuffer_0 == NULL) {
	localTerminationProcedure(
	    "ERROR: realloc failed.  probably not enough memory for allocation."
	);
    }

	/* if array holds more than 1 non-zero struct, chop off the */
	/* last non-zero struct if it is valid at secondsSince1996Stop. */
	/* this is to prevent the return of overlapping data on a */
	/* series of calls to getStructs. */
    nonzeroStructCount = structCount - 1;
    if (nonzeroStructCount > 1) {
	int zeroedStructCount;

	zeroedStructCount = zeroLastStructIfValidAtStopTime(pStructBuffer_0,
			structName, secondsSince1996Stop, (structCount - 1));
	nonzeroStructCount -= zeroedStructCount;
    }

    if (overlapStructCount > 0) {
	pOverlapStructs = pOverlapStructs_0;
	pStructBuffer_0 = integrateOverlapStructs(structName, pStructBuffer_0,
		nonzeroStructCount, pOverlapStructs, overlapStructCount);
    }
    free(pOverlapStructs_0);

    return pStructBuffer_0;
}

static int zeroLastStructIfValidAtStopTime(char *pStructs, char *structName,
				uint32 stopTime, int nonzeroStructCount) {
    int structSize, i;
    uint32 lastSecondOfStructsValidity;

    structSize = getStructSizeFromName(structName);
    pStructs += (nonzeroStructCount - 1) * structSize;

	/* for those data types with stop time stamps... */
    lastSecondOfStructsValidity = getStopTimeStamp(pStructs, structName);
	/* for periodic data types... */
    if (lastSecondOfStructsValidity == 0)
	lastSecondOfStructsValidity = getTimeStamp(pStructs, structName) +
			getDataTypePeriodFromStructName(structName) - 1;
    if (stopTime <= lastSecondOfStructsValidity) {
	    /* zero this struct in the array, prevents data overlap over */
	    /* a series of calls to getStructs.				 */
	for (i = 0; i < structSize; i++) *pStructs++ = 0; /* zero out struct */
	return 1;	/* return number of zeroed structs */
    }
    else return 0;	/* return number of zeroed structs */
}

static int countNonzeroStructsInArray(char *pStructs, char *structName) {
    int count, structSize;
    uint32 timeStamp;

    structSize = getStructSizeFromName(structName);

    count = 0;
    timeStamp = getTimeStamp(pStructs, structName);
    while (timeStamp != 0) {
	count++;
	pStructs += structSize;
	timeStamp = getTimeStamp(pStructs, structName);
    }
    return count;
}

static void localTerminationProcedure(char *errorMessage) {
    fprintf(stderr, "%s\n", errorMessage);
    if (getAppTerminatePermission()) {
	fprintf(stderr, "\n\tExiting.\n\n");
	exit(1);
    }
    else fprintf(stderr, "\n\tSERIOUS ERROR!\n\n");
}

static char *integrateOverlapStructs(char *structName, char *pStructs,
			int nonOverlapStructCount, char *pOverlapStructs,
						int overlapStructCount) {
    char *pOldLoc, *pNewLoc;
    int structSize, dataSize, i;

    structSize = getStructSizeFromName(structName);
    dataSize = structSize * (nonOverlapStructCount + overlapStructCount + 1);
    pStructs = realloc(pStructs, dataSize);

	/* point to last struct (the zeroed struct, pre-integration) */
    pOldLoc = pStructs + (structSize * nonOverlapStructCount);
    pNewLoc = pOldLoc + (structSize * overlapStructCount);

	/* move the zeroed struct to new location in bigger array */
    memcpy(pNewLoc, pOldLoc, structSize);
    pOldLoc -= structSize;
    pNewLoc -= structSize;

	/* shift structs in old array location up to make room for overlaps */
	/* start at latest time overlap and work down */
    for (i = overlapStructCount - 1; i >= 0; i--) {
	char *pOverlap;
	uint32 overlapTimeStamp;

	    /* get to the next overlap struct and get its time stamp */
	pOverlap = pOverlapStructs + (i * structSize);
	overlapTimeStamp = getTimeStamp(pOverlap, structName);

	    /* move structs to new location until overlap should be placed */
	while (getTimeStamp(pOldLoc, structName) > overlapTimeStamp) {
	    memcpy(pNewLoc, pOldLoc, structSize);
	    pOldLoc -= structSize;
	    pNewLoc -= structSize;
	}

	    /* put in the overlap struct */
	memcpy(pNewLoc, pOverlap, structSize);
	pNewLoc -= structSize;
	    /* note that pOldLoc is not decremented.  nothing */
	    /* has been moved from the old location here. */
    }

    return pStructs;
}


/*

getWorkingStartTime is needed in order to determine if the secondsSince1996
falls within the time that the last non-zeroed struct time stamped
prior to secondsSince1996 was valid.  in other words, search backward from
secondsSince1996 to find the first struct containing real data and determine
if secondsSince1996 falls within the time that struct was valid.

usually this is just a case of going back a second less than one cycle
and seeing if any structs are time stamped between 
(secondsSince1996 - period + 1) and secondsSince1996.  if so, this is
the struct valid at secondsSince1996, and we want to make sure we
include it in any set of structs we are getting.

unfortunately, there are also some non-periodic structs which have
both a start and stop time in the struct.  in these cases we have to
back up to the last non-zero struct and check the stop time to see
if it was valid at secondsSince1996.

*/

    /* local subroutines for getWorkingStartTime */
static uint32 getStartForStructWithStopTime(uint32 secondsSince1996,
							char *structName);
static void getBeginEndSecsOfStructPriorTo(uint32 *startAndStop,
		uint32 secondsSince1996, char *dataFilePath, char *structName);


static uint32 getWorkingStartTime(uint32 secondsSince1996, char *structName) {
    if (strstr(structName, "Cris") != NULL) {
	if (strcmp(structName, "BrowseCris") == 0)
	    return secondsSince1996 - CRIS_BROWSE_SAMPLE_PERIOD + 1;
	else if (strcmp(structName, "L1CrisCommandEcho") == 0)
	    return secondsSince1996 - CRIS_COMMAND_ECHO_SAMPLE_PERIOD + 1;
	else if (strcmp(structName, "L1CrisCommandTable") == 0) return
		getStartForStructWithStopTime(secondsSince1996, structName);
	else if (strcmp(structName, "L1CrisCommandTableA") == 0) return
		getStartForStructWithStopTime(secondsSince1996, structName);
	else if (strcmp(structName, "L1CrisDiagnosticEvent") == 0) return
		getStartForStructWithStopTime(secondsSince1996, structName);
	else if (strcmp(structName, "L1CrisEventCycle") == 0)
	    return secondsSince1996 - CRIS_EVENT_CYCLE_SAMPLE_PERIOD + 1;
	else if (strcmp(structName, "L1CrisHighPriorityRate") == 0)
	    return secondsSince1996 - CRIS_HIGH_PRIORITY_RATE_SAMPLE_PERIOD + 1;
	else if (strcmp(structName, "L1CrisHskp") == 0)
	    return secondsSince1996 - CRIS_HOUSEKEEPING_SAMPLE_PERIOD + 1;
	else if (strcmp(structName, "L1CrisLowPriorityRate") == 0)
	    return secondsSince1996 - CRIS_LOW_PRIORITY_RATE_SAMPLE_PERIOD + 1;
	else if (strcmp(structName, "L1CrisSubset") == 0)
	    return secondsSince1996 - CRIS_SUBSET_SAMPLE_PERIOD + 1;
	else if (strcmp(structName, "L1CrisSummary") == 0)
	    return secondsSince1996 - CRIS_SUMMARY_SAMPLE_PERIOD + 1;
    }
    else {
	if (strcmp(structName, "BrowseSis") == 0)
	    return secondsSince1996 - SIS_BROWSE_SAMPLE_PERIOD + 1;
	else if (strcmp(structName, "L1SisCommandEcho") == 0)
	    return secondsSince1996 - SIS_COMMAND_ECHO_SAMPLE_PERIOD + 1;
	else if (strcmp(structName, "L1SisCommandTable") == 0) return
		getStartForStructWithStopTime(secondsSince1996, structName);
	else if (strcmp(structName, "L1SisDacOffset0") == 0)
	    return secondsSince1996 - SIS_DAC_OFFSET_0_SAMPLE_PERIOD + 1;
	else if (strcmp(structName, "L1SisDacOffset1") == 0)
	    return secondsSince1996 - SIS_DAC_OFFSET_1_SAMPLE_PERIOD + 1;
	else if (strcmp(structName, "L1SisDiagnosticEvent") == 0) return
		getStartForStructWithStopTime(secondsSince1996, structName);
	else if (strcmp(structName, "L1SisEventCycle") == 0)
	    return secondsSince1996 - SIS_EVENT_CYCLE_SAMPLE_PERIOD + 1;
	else if (strcmp(structName, "L1SisHighPriorityRate") == 0)
	    return secondsSince1996 - SIS_HIGH_PRIORITY_RATE_SAMPLE_PERIOD + 1;
	else if (strcmp(structName, "L1SisHskp") == 0)
	    return secondsSince1996 - SIS_HOUSEKEEPING_SAMPLE_PERIOD + 1;
	else if (strcmp(structName, "L1SisLowPriorityRate") == 0)
	    return secondsSince1996 - SIS_LOW_PRIORITY_RATE_SAMPLE_PERIOD + 1;
	else if (strcmp(structName, "L1SisSubset") == 0)
	    return secondsSince1996 - SIS_SUBSET_SAMPLE_PERIOD + 1;
	else if (strcmp(structName, "L1SisSummary") == 0)
	    return secondsSince1996 - SIS_SUMMARY_SAMPLE_PERIOD + 1;
    }

	/* if we got here, did not recognize the structure name */
    fprintf(stderr, "getWorkingStartTime: ERROR\n");
    fprintf(stderr, "  Struct name ->%s<- not recognized.  Exiting.\n\n",
								structName);
    exit(0);
}

    /* in explanation of the strangely named subroutine, structs which */
    /* are acyclic and have stop times associated with their intervals */
    /* of validity instead of a constant period, need to be treated in */
    /* a special manner to see if secondsSince1996 falls within the */
    /* valid interval of the structure previous to the request time, */
    /* secondsSince1996... */

static uint32 getStartForStructWithStopTime(uint32 secondsSince1996,
							char *structName) {
    uint32 startAndStopOfPreviousStruct[2];
    uint32 previousStart, previousStop;
    char path[256];
    struct tm *pTimeStruct;
    char instrumentID;
    char suffix[10];

    pTimeStruct = getDateFromSecondsSince1996(secondsSince1996);
    instrumentID = getInstrumentIDFromStructName(structName);
    getDataFileSuffixFromStructName(suffix, structName);
    generateDataFilePath(path, pTimeStruct, instrumentID, suffix);

    getBeginEndSecsOfStructPriorTo(startAndStopOfPreviousStruct,
					secondsSince1996, path, structName);
    previousStart = startAndStopOfPreviousStruct[0];
    previousStop = startAndStopOfPreviousStruct[1];


	/* if getBeginEndSecsOfStructPriorTo returned zeros, it means no */
	/* previous non-zero struct was found.  stick with looking for */
	/* start tags greater than secondsSince1996. */
    if (previousStop == 0) return secondsSince1996;

	/* else if secondsSince1996 occurs while the previous struct was */
	/* valid, return the previous structs start time as the "working */
	/* start time." */
    else if (secondsSince1996 <= previousStop) return previousStart;

	/* else previous struct was not valid at secondsSince1996, so */
	/* again, stick with looking for start tags greater than */
	/* secondsSince1996. */
    else return secondsSince1996;
}


    /* only command tables and diagnostic events should be calling */
    /* this routine. */
static void getBeginEndSecsOfStructPriorTo(uint32 *startAndStop,
		uint32 secondsSince1996, char *dataFilePath, char *structName) {
    int ct = 0, de = 0;
    int day1996, index, fd;
    struct tm *pTimeStruct;
    uint32 dayStartSecond, secondsSince12AM, stamp;
    char *pBuffer;

    if (strstr(structName, "CommandTable") != NULL) ct = 1;
    else if (strstr(structName, "DiagnosticEvent") != NULL) de = 1;
    else {
	    /* if we got here, invalid call */
	fprintf(stderr, "getStructSizeFromName: ERROR\n");
	fprintf(stderr, "  Invalid call to getBeginEndSecsOfStructPriorTo.\n");
	fprintf(stderr, "  Struct name ->%s<- not recognized.  Exiting.\n\n",
								structName);
	exit(0);
    }

    day1996 = getDay1996FromSecondsSince1996(secondsSince1996);
    pTimeStruct = getDateFromDay1996(day1996);
    dayStartSecond = getSecondsSince1996FromDate(pTimeStruct);
    secondsSince12AM = secondsSince1996 - dayStartSecond;

    if (!(fileExists(dataFilePath))) {
	    /* file does not exist */
	fprintf(stderr, "WARNING: File %s does not exist.\n", dataFilePath);
	fprintf(stderr, "No data that day? Checking previous days events.\n\n");
	index = -1;
    }
    else {
	    /* file does exist */
	int structCount, structSize, hasHeader;

	hasHeader = 1;
	structSize = getStructSizeFromName(structName);
	pBuffer = malloc(structSize);
	structCount = getFilesStructCount(dataFilePath, structSize, hasHeader);
	    /* initial guess */
	index = (int) (structCount * (secondsSince12AM / 86400.));

	fd = open(dataFilePath, O_RDONLY | O_BINARY, 0);    /* open to read */
	lseek(fd, HEADER_BYTES + (index * structSize), 0);
	readFromFile(fd, pBuffer, structSize);
	stamp = getTimeStamp(pBuffer, structName);

	    /* step forward past zeroed structs */
	while (stamp == 0) {
	    index++;
	    if (index >= structCount) break;
	    lseek(fd, HEADER_BYTES + (index * structSize), 0);
	    readFromFile(fd, pBuffer, structSize);
	    stamp = getTimeStamp(pBuffer, structName);
	}

	if ((stamp != 0) && (stamp <= secondsSince1996)) {
		/* step forward past structs occuring before secondsSince1996 */
	    while (stamp <= secondsSince1996) {
		index++;
		if (index >= structCount) break;
		lseek(fd, HEADER_BYTES + (index * structSize), 0);
		readFromFile(fd, pBuffer, structSize);
		stamp = getTimeStamp(pBuffer, structName);
	    }

		/* step back one struct */
	    index--;
	    lseek(fd, HEADER_BYTES + (index * structSize), 0);
	    readFromFile(fd, pBuffer, structSize);
	    stamp = getTimeStamp(pBuffer, structName);
		/* if stamp is 0, step back till finding non-zero stamp */
	    while (stamp ==0) {
		index--;
		if (index < 0) break;
		lseek(fd, HEADER_BYTES + (index * structSize), 0);
		readFromFile(fd, pBuffer, structSize);        
		stamp = getTimeStamp(pBuffer, structName);
	    }
	}
	else {
		/* step back before structs occuring later than */
		/* secondsSince1996.  also ignore stamps equal to zero. */
	    while ((stamp > secondsSince1996) || (stamp == 0)) {
		index--;
		if (index < 0) break;
		lseek(fd, HEADER_BYTES + (index * structSize), 0);
		readFromFile(fd, pBuffer, structSize);
		stamp = getTimeStamp(pBuffer, structName);
	    }
	}

	if (index >= 0) {
		/* prior non-zero struct lies in this day's structs (yeah!) */
	    lseek(fd, HEADER_BYTES + (index * structSize), 0);
	    readFromFile(fd, pBuffer, structSize);
	    startAndStop[0] = getTimeStamp(pBuffer, structName);
	    startAndStop[1] = getStopTimeStamp(pBuffer, structName);
	    close(fd);
	    free(pBuffer);
	    return;
	}
	else {
		/* ugh! deal with looking in previous day's data file */
	    char newPath[256];
	    char instrumentID;
	    char suffix[10];

	    close(fd);

		/* determine path to previous day's file */
	    day1996--;
	    pTimeStruct = getDateFromDay1996(day1996);
	    instrumentID = getInstrumentIDFromStructName(structName);
	    getDataFileSuffixFromStructName(suffix, structName);
	    generateDataFilePath(newPath, pTimeStruct, instrumentID, suffix);

	    if (!(fileExists(newPath))) {
		startAndStop[0] = 0L;
		startAndStop[1] = 0L;
		free(pBuffer);
		return;
	    }

		/* look for last non-zero struct in previous day's file */
	    structCount = getFilesStructCount(newPath, structSize, hasHeader);
	    index = structCount - 1;
	    fd = open(newPath, O_RDONLY | O_BINARY, 0);
	    lseek(fd, HEADER_BYTES + (index * structSize), 0);
	    readFromFile(fd, pBuffer, structSize);
	    stamp = getTimeStamp(pBuffer, structName);
		/* step back past zeroed structs */
	    while (stamp == 0) {
		index--;
		if (index < 0) break;
		lseek(fd, HEADER_BYTES + (index * structSize), 0);
		readFromFile(fd, pBuffer, structSize);
		stamp = getTimeStamp(pBuffer, structName);
	    }

	    if (index >= 0) {
		lseek(fd, HEADER_BYTES + (index * structSize), 0);
		readFromFile(fd, pBuffer, structSize);
		startAndStop[0] = getTimeStamp(pBuffer, structName);
		startAndStop[1] = getStopTimeStamp(pBuffer, structName);
		close(fd);
		free(pBuffer);
		return;
	    }
	    else {
		    /* didn't find a non-zero struct */
		startAndStop[0] = 0L;
		startAndStop[1] = 0L;
		close(fd);
		free(pBuffer);
		return;
	    }
	}
    }
}

    /* local utility routines for getSubsetOfDaysStructs */
static void getDataBlockStartAndStopNumbers(int day1996, int fd,
		char *structName, uint32 getStampsAfter, uint32 getStampsUpTo,
		int dataBlocksInFile, int *startAndStopStructNumbers);


static void *getSubsetOfDaysStructs(int day1996, char *path, char *structName,
				uint32 getStampsAfter, uint32 getStampsUpTo) {
    char *pStructArray_0, *pNextStructInArray, *dataBuffer, *pIntoDataBuffer;
    int structSize, hasHeader, structsInFileCount, fd, i;
    int firstValidDataBlockNumber, lastValidDataBlockNumber, maxValidDataBlocks;
    int startAndStopNumbers[2];
    int goodStructCount, doneReading;
    uint32 timeStamp;

    structSize = getStructSizeFromName(structName);
    hasHeader = 1;
    structsInFileCount = getFilesStructCount(path, structSize, hasHeader);

	/* if no structs were found, return pointer to single terminal struct */
    if (structsInFileCount == 0) {
	pStructArray_0 = (char *) calloc(1, structSize);
	memset(pStructArray_0, (char) 0, structSize);  /* zero out struct */
	return pStructArray_0;
    }

	/* open the data file for reading	*/
    if ((fd = open(path, O_RDONLY | O_BINARY, 0)) == -1) {
	    /* can't open the file for some reason */
	char CANT_OPEN_ERROR_MSG[] = "Unable to open file %s\n\nExiting.\n\n";
	perror("getSubsetOfDaysStructs\n\t");
	fprintf(stderr, CANT_OPEN_ERROR_MSG, path);
	exit(1);
    }

	/* find the first and last data structs in the file with a time	*/
	/*     stamp later than getStampsAfter and before or equal to	*/
	/*     getStampsUpTo.						*/

    getDataBlockStartAndStopNumbers(day1996, fd, structName, getStampsAfter,
			getStampsUpTo, structsInFileCount, startAndStopNumbers);
    firstValidDataBlockNumber = startAndStopNumbers[0];
    lastValidDataBlockNumber = startAndStopNumbers[1];

	/* if no valid structs were found, return pointer to single	*/
	/*	terminal struct						*/
    if (firstValidDataBlockNumber < 0) {
	close(fd);
	pStructArray_0 = (char *) calloc(1, structSize);
	memset(pStructArray_0, (char) 0, structSize);  /* zero out struct */
	return pStructArray_0;
    }

    maxValidDataBlocks = lastValidDataBlockNumber+1 - firstValidDataBlockNumber;

	/* allocate memory for structs (+1 for terminal struct) */
	/*     dataBuffer is used to read everything that is possibly	*/
	/*     valid in one read.  pStructArray_0 is the array	*/
	/*     which will eventually be passed back after copying all	*/
	/*     valid structs from dataBuffer.				*/
    dataBuffer = (char *) calloc((maxValidDataBlocks + 1), structSize);
    pStructArray_0 = (char *) calloc((maxValidDataBlocks + 1), structSize);

	/* go to first valid data struct in file */
    lseek(fd, (HEADER_BYTES + (firstValidDataBlockNumber * structSize)), 0);
    readFromFile(fd, dataBuffer, (maxValidDataBlocks * structSize));
    close(fd);

    goodStructCount = 0;
    doneReading = 0;
    pIntoDataBuffer = dataBuffer;
    pNextStructInArray = pStructArray_0;
    for (i = 0; i < maxValidDataBlocks; i++) {
	timeStamp = getTimeStamp(pIntoDataBuffer, structName);
	if ((timeStamp <= getStampsAfter) || (timeStamp == 0xFFFFFFFF)) {
	    /* invalid struct (zeroed struct meaning no data		*/
	    /*     is archived in this slot or struct filled with	*/
	    /*     (char) -1s, meaning data is in the overlap file	*/
	    /*     here).  no copy.					*/
	}
	else {
	    memcpy(pNextStructInArray, pIntoDataBuffer, structSize);
	    pNextStructInArray += structSize;
	    goodStructCount++;
	}
	pIntoDataBuffer += structSize;
    }
    free(dataBuffer);

	/* zero the terminal struct */
    memset(pNextStructInArray, (char) 0, structSize);

	/* cut struct array down to size */
    pStructArray_0 = realloc(pStructArray_0,
				(goodStructCount + 1) * structSize);
    return pStructArray_0;
}


static void getDataBlockStartAndStopNumbers(int day1996, int fd,
		char *structName, uint32 getStampsAfter, uint32 getStampsUpTo,
		int dataBlocksInFile, int *startAndStopStructNumbers) {
    struct tm *pTm;
    char *buffer;
    uint32 secondsSince1996AtDayStart, timeStamp;
    int secondsIntoDay, dataPeriod, dataBlockNumberInFile, dataBlockSize;

    pTm = getDateFromDay1996(day1996);
	/* make sure we're getting the secondsSince1996 at the day's start */
    pTm->tm_hour = 0;
    pTm->tm_min = 0;
    pTm->tm_sec = 0;
    secondsSince1996AtDayStart = getSecondsSince1996FromDate(pTm);
    secondsIntoDay = (int) (getStampsAfter - secondsSince1996AtDayStart);
    if (secondsIntoDay < 0) secondsIntoDay = 0;

    dataPeriod = getDataTypePeriodFromStructName(structName);
    if (dataPeriod > 0) dataBlockNumberInFile = secondsIntoDay / dataPeriod;
    else dataBlockNumberInFile = 0;

    dataBlockSize = getStructSizeFromName(structName);
    buffer = malloc(dataBlockSize);

	/* step forward through file from here to get right data block */
    lseek(fd, (HEADER_BYTES + (dataBlockNumberInFile * dataBlockSize)), 0);
    for (; dataBlockNumberInFile < dataBlocksInFile; dataBlockNumberInFile++) {
	readFromFile(fd, buffer, dataBlockSize);
	timeStamp = getTimeStamp(buffer, structName);
	if (timeStamp > getStampsAfter) break;
    }

	/* found no valid data block */
    if (dataBlockNumberInFile == dataBlocksInFile) {
	startAndStopStructNumbers[0] = -1;
	startAndStopStructNumbers[1] = -1;
	free(buffer);
	return;
    }
    else startAndStopStructNumbers[0] = dataBlockNumberInFile;

    secondsIntoDay = (int) (getStampsUpTo - secondsSince1996AtDayStart);
    if (secondsIntoDay >= 86400) secondsIntoDay = 86399;
    if (dataPeriod > 0) dataBlockNumberInFile = secondsIntoDay / dataPeriod;
    else dataBlockNumberInFile = 0;

    lseek(fd, (HEADER_BYTES + (dataBlockNumberInFile * dataBlockSize)), 0);
    for (; dataBlockNumberInFile < dataBlocksInFile; dataBlockNumberInFile++) {
	readFromFile(fd, buffer, dataBlockSize);
	timeStamp = getTimeStamp(buffer, structName);
	if (timeStamp > getStampsUpTo) break;
    }
	/* we went 1 block past the last struct in the time interval	*/
    dataBlockNumberInFile--;

    free(buffer);

    startAndStopStructNumbers[1] = dataBlockNumberInFile;
    return;
}
