#include <stdio.h>
#include <time.h>
#include <string.h>
#include "hdfi.h"

#include "porting_defines.h"
#include "utilities.h"
#include "crisSisPutUtilities.h"

/*

version 1.2:
    added passed parameter char instrumentID = 'c' or 's' for CRIS and
    SIS respectively, so same code can be used for both.

determineDataFilePathAndCreateDirsAsNeeded takes the number of seconds
since 00:00:00 1/1/96, a data type file suffix, and returns the path
for the data file (of the form .../s1996.01/s1996.01.02/s1996.01.02.evt)
and also the interval number that secondsSinceMidnight fits into in
the blocks that cover the day in question.

*/

void determineDataFilePathAndCreateDirsAsNeeded(uint32 secondsSince1996,
				const char *dataFileSuffix, char *path,
				int *pIntervalNumber, char instrumentID) {
    struct tm *pTime_str;
    int year, month, mday;
    char dd = DIRECTORY_DELIMITER_aceware_level1; /* see porting_defines.h */
    char dd_str[] = { DIRECTORY_DELIMITER_aceware_level1, '\0' };
    char yearMonthString[8], dateString[10];
    char *BASE_DIRECTORY, *pChar, *pToken;
    char baseDirShort[256], buffer[256];

    if ((instrumentID == 'C') || (instrumentID == 'c')) {	/* CRIS */
	instrumentID = 'c';
	BASE_DIRECTORY = getenv("L1_CRIS_DATA_BASE_DIRECTORY");
	if (BASE_DIRECTORY == NULL) {
	    fprintf(stderr, "The environment variable ");
	    fprintf(stderr, "L1_CRIS_DATA_BASE_DIRECTORY");
	    fprintf(stderr, "\nhas not been defined.\n\nExiting.\n\n");
	    exit(1);
	}
    }
    else if ((instrumentID == 'S') || (instrumentID == 's')) {       /* SIS */
	instrumentID = 's';
	BASE_DIRECTORY = getenv("L1_SIS_DATA_BASE_DIRECTORY");
	if (BASE_DIRECTORY == NULL) {
	    fprintf(stderr, "The environment variable ");
	    fprintf(stderr, "L1_SIS_DATA_BASE_DIRECTORY");
	    fprintf(stderr, "\nhas not been defined.\n\nExiting.\n\n");
	    exit(1);
	}
    }
    else {
	fprintf(stderr, "Invalid instrument ID: %c.\n\nExiting.\n\n",
								instrumentID);
	exit(1);
    }
    if (!directoryExists(BASE_DIRECTORY)) {
	fprintf(stderr, "The base directory %s does not exist.\n\nExiting.\n\n",
								BASE_DIRECTORY);
	exit(1);
    }

	/* use directories after the "data" directory in the environment */
	/* variable L1_[CRIS|SIS]_BASE_DIRECTORY as a file identifier to */
	/* be prepended to the data file name (baseDirShort refers to the */
	/* base directory, shortened by chopping of everything up to "data") */
    strcpy(buffer, BASE_DIRECTORY);
    pChar = buffer;
    strcpy(baseDirShort, "");
    while ((pToken = strtok(pChar, dd_str))) {
	pChar = NULL;
	if (!strcmp(pToken, "data")) {	/* if (token == "data") */
	    while ((pToken = strtok(pChar, dd_str))) {
		if (!strcmp(pToken, "data")) {
		    fprintf(stderr,
			"Directory \"data\" appears multiple times in base");
		    fprintf(stderr, "directory \"%s\"\n\nExiting.\n\n",
								BASE_DIRECTORY);
		    exit(1);
		}
		if (!strcmp(pToken, "")) break;		/* ignore nothing */
		if (*pToken == ' ') break;		/* ignore spaces */
		if (*pToken == '	') break;	/* ignore tabs */
		strcat(baseDirShort, pToken);
		strcat(baseDirShort, ".");
	    }
	    break;
	}
    }
    if (!strcmp(baseDirShort, "")) {
	fprintf(stderr, "\"data\" directory and/or its subdirectories not");
	fprintf(stderr, " found in base directory \"%s\"\n\nExiting.\n\n",
								BASE_DIRECTORY);
	exit(1);
    }

	/* pIntervalNumber points to a location to store a value which */
	/* is used later.  this is placed in a block just to separate it. */
    {
	int day1996;
	uint32 zerothSecondOfThisDay, secondsSinceMidnight;

	day1996 = getDay1996FromSecondsSince1996(secondsSince1996);
	zerothSecondOfThisDay = getSecondsSince1996FromDay1996(day1996);
	secondsSinceMidnight = secondsSince1996 - zerothSecondOfThisDay;
	*pIntervalNumber = secondsSinceMidnight / DATA_BLOCK_INTERVAL;
    }
 
    pTime_str = getDateFromSecondsSince1996(secondsSince1996);
    year = pTime_str->tm_year + 1900;
    month = pTime_str->tm_mon + 1;
    mday = pTime_str->tm_mday;

    sprintf(yearMonthString, "%4.4u.%2.2u", year, month);
    sprintf(path, "%s%c%s", BASE_DIRECTORY, dd, yearMonthString);
    if (!directoryExists(path)) {
	/* fprintf(stderr, "directory %s does not exist\n", path); */
	createDirectory(path);
    }
 
    sprintf(dateString, "%4.4u.%2.2u.%2.2u", year, month, mday);
    sprintf(path, "%s%c%s%c%s",BASE_DIRECTORY,dd,yearMonthString,dd,dateString);
    if (!directoryExists(path)) {
	/* fprintf(stderr, "directory %s does not exist\n", path); */
	createDirectory(path);
    }
 
    sprintf(path, "%s%c%s%c%s%c%s%s.%s", BASE_DIRECTORY, dd, yearMonthString,
		dd, dateString, dd, baseDirShort, dateString, dataFileSuffix);
    /* printf("path = %s\n", path); */
}
